@@ -18,19 +18,19 @@
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #
-# Translated to perk/Tk by Slaven Rezic <eserte@cs.tu-berlin.de>.
+# Translated to perl/Tk by Slaven Rezic <slaven@rezic.de>.
 #
 
 #----------------------------------------------------------------------
 #
-#		      F I L E   D I A L O G
+#                     F I L E   D I A L O G
 #
 #----------------------------------------------------------------------
 # tkFDialog --
 #
-#	Implements the TK file selection dialog. This dialog is used when
-#	the tk_strictMotif flag is set to false. This procedure shouldn't
-#	be called directly. Call tk_getOpenFile or tk_getSaveFile instead.
+#       Implements the TK file selection dialog. This dialog is used when
+#       the tk_strictMotif flag is set to false. This procedure shouldn't
+#       be called directly. Call tk_getOpenFile or tk_getSaveFile instead.
 #
 
 package Tk::FBox;
@@ -39,21 +39,28 @@ require Tk::Toplevel;
 use strict;
 use vars qw($VERSION $updirImage $folderImage $fileImage);
 
-$VERSION = '4.009'; # $Id: //depot/Tkutf8/Tk/FBox.pm#9 $
+#$VERSION = sprintf '4.%03d', q$Revision: #18 $ =~ /\D(\d+)\s*$/;
+$VERSION = '4.020';
 
 use base qw(Tk::Toplevel);
 
 Construct Tk::Widget 'FBox';
 
-my $selectFilePath;
-my $selectFile;
-
 sub import {
     if (defined $_[1] and $_[1] eq 'as_default') {
 	local $^W = 0;
 	package Tk;
-	*FDialog      = \&Tk::FBox::FDialog;
-	*MotifFDialog = \&Tk::FBox::FDialog;
+	if ($Tk::VERSION < 804) {
+	    *FDialog      = \&Tk::FBox::FDialog;
+	    *MotifFDialog = \&Tk::FBox::FDialog;
+	} else {
+	    *tk_getOpenFile = sub {
+		Tk::FBox::FDialog("tk_getOpenFile", @_);
+	    };
+	    *tk_getSaveFile = sub {
+		Tk::FBox::FDialog("tk_getSaveFile", @_);
+	    };
+	}
     }
 }
 
@@ -72,17 +79,18 @@ sub Populate {
 
     $w->SUPER::Populate($args);
 
+    $w->{'encoding'} = $w->getEncoding('iso_8859_1');
+
     # f1: the frame with the directory option menu
     my $f1 = $w->Frame;
     my $lab = $f1->Label(-text => 'Directory:', -underline => 0);
-    $w->{'selectPath'} = '.';
     $w->{'dirMenu'} = my $dirMenu =
       $f1->Optionmenu(-variable => \$w->{'selectPath'},
 		      -textvariable => \$w->{'selectPath'},
 		      -command => ['SetPath', $w]);
     my $upBtn = $f1->Button;
     if (!defined $updirImage->{$w->MainWindow}) {
-        $updirImage->{$w->MainWindow} = $w->Bitmap(-data => <<EOF);
+	$updirImage->{$w->MainWindow} = $w->Bitmap(-data => <<EOF);
 #define updir_width 28
 #define updir_height 16
 static char updir_bits[] = {
@@ -101,12 +109,13 @@ EOF
     $dirMenu->pack(-expand => 'yes', -fill => 'both', -padx => 4);
 
     $w->{'icons'} = my $icons =
-      $w->IconList(-browsecmd => ['ListBrowse', $w],
-		   -command   => ['ListInvoke', $w],
+      $w->IconList(-command => ['OkCmd', $w, 'iconlist'],
 		  );
+    $icons->bind('<<ListboxSelect>>' => [$w, 'ListBrowse']);
 
     # f2: the frame with the OK button and the "file name" field
     my $f2 = $w->Frame(-bd => 0);
+#XXX File name => File names if multiple
     my $f2_lab = $f2->Label(-text => 'File name:', -anchor => 'e',
 			    -width => 14, -underline => 5, -pady => 0);
     $w->{'ent'} = my $ent = $f2->Entry;
@@ -134,6 +143,7 @@ EOF
        -relief => $f2_lab->cget(-relief),
        -padx => $f2_lab->cget(-padx),
        -pady => $f2_lab->cget(-pady),
+       -takefocus => 0,
       );
     $typeMenuLab->bindtags([$typeMenuLab, 'Label',
 			    $typeMenuLab->toplevel, 'all']);
@@ -185,9 +195,9 @@ EOF
 
     $w->bind('<Alt-d>',[$dirMenu,'focus']);
     $w->bind('<Alt-t>',sub  {
-                             if ($typeMenuBtn->cget(-state) eq 'normal') {
-                             $typeMenuBtn->focus;
-                             } });
+			     if ($typeMenuBtn->cget(-state) eq 'normal') {
+			     $typeMenuBtn->focus;
+			     } });
     $w->bind('<Alt-n>',[$ent,'focus']);
     $w->bind('<KeyPress-Escape>',[$cancelBtn,'invoke']);
     $w->bind('<Alt-c>',[$cancelBtn,'invoke']);
@@ -204,17 +214,18 @@ EOF
     $w->SetPath(_cwd());
 
     $w->ConfigSpecs(-defaultextension => ['PASSIVE', undef, undef, undef],
-                    -filetypes        => ['PASSIVE', undef, undef, undef],
-                    -initialdir       => ['PASSIVE', undef, undef, undef],
-                    -initialfile      => ['PASSIVE', undef, undef, undef],
-#		    -sortcmd          => ['PASSIVE', undef, undef, sub { lc($a) cmp lc($b) }],
+		    -filetypes        => ['PASSIVE', undef, undef, undef],
+		    -initialdir       => ['PASSIVE', undef, undef, undef],
+		    -initialfile      => ['PASSIVE', undef, undef, undef],
+#                   -sortcmd          => ['PASSIVE', undef, undef, sub { lc($a) cmp lc($b) }],
 		    -sortcmd          => ['PASSIVE', undef, undef, sub { lc($_[0]) cmp lc($_[1]) }],
-                    -title            => ['PASSIVE', undef, undef, undef],
-                    -type             => ['PASSIVE', undef, undef, 'open'],
-                    -filter           => ['PASSIVE', undef, undef, '*'],
-                    -force            => ['PASSIVE', undef, undef, 0],
-                    'DEFAULT'         => [$icons],
-                   );
+		    -title            => ['PASSIVE', undef, undef, undef],
+		    -type             => ['PASSIVE', undef, undef, 'open'],
+		    -filter           => ['PASSIVE', undef, undef, '*'],
+		    -force            => ['PASSIVE', undef, undef, 0],
+		    -multiple         => ['PASSIVE', undef, undef, 0],
+		    'DEFAULT'         => [$icons],
+		   );
     # So-far-failed attempt to break reference loops ...
     $w->_OnDestroy(qw(dirMenu icons typeMenuLab typeMenuBtn okBtn ent updateId));
     $w;
@@ -226,7 +237,16 @@ sub Show {
 
     $w->configure(@_);
 
-    $w->transient($w->Parent);
+    # Dialog boxes should be transient with respect to their parent,
+    # so that they will always stay on top of their parent window.  However,
+    # some window managers will create the window as withdrawn if the parent
+    # window is withdrawn or iconified.  Combined with the grab we put on the
+    # window, this can hang the entire application.  Therefore we only make
+    # the dialog transient if the parent is viewable.
+
+    if (Tk::Exists($w->Parent) && $w->Parent->viewable) {
+	$w->transient($w->Parent);
+    }
 
     # set the default directory and selection according to the -initial
     # settings
@@ -243,6 +263,15 @@ sub Show {
 	$w->{'selectFile'} = $w->cget(-initialfile);
     }
 
+    # Set -multiple to a one or zero value (not other boolean types
+    # like "yes") so we can use it in tests more easily.
+    if ($w->cget('-type') ne 'open') {
+	$w->configure(-multiple => 0);
+    } else {
+	$w->configure(-multiple => !!$w->cget('-multiple'));
+    }
+    $w->{'icons'}->configure(-multiple => $w->cget('-multiple'));
+
     # Initialize the file types menu
     my $typeMenuBtn = $w->{'typeMenuBtn'};
     my $typeMenuLab = $w->{'typeMenuLab'};
@@ -262,32 +291,43 @@ sub Show {
 	$typeMenuBtn->configure(-state => 'normal');
 	$typeMenuLab->configure(-state => 'normal');
     } else {
-	$w->configure(-filter => '*');
+#XXX    $w->configure(-filter => '*');
 	$typeMenuBtn->configure(-state => 'disabled',
 				-takefocus => 0);
 	$typeMenuLab->configure(-state => 'disabled');
     }
     $w->UpdateWhenIdle;
 
+    {
+	my $title = $w->cget(-title);
+	if (!defined $title) {
+	    my $type = $w->cget(-type);
+	    $title = ($type eq 'dir') ? 'Choose Directory'
+                     : ($type eq 'save') ? 'Save As' : 'Open';
+	}
+	$w->title($title);
+    }
+
     # Withdraw the window, then update all the geometry information
     # so we know how big it wants to be, then center the window in the
     # display and de-iconify it.
     $w->withdraw;
     $w->idletasks;
-    my $x = int($w->screenwidth / 2 - $w->reqwidth / 2 - $w->parent->vrootx);
-    my $y = int($w->screenheight / 2 - $w->reqheight / 2 - $w->parent->vrooty);
-    $w->geometry("+$x+$y");
+    if (0)
+     {
+      #XXX use Tk::Wm::Popup? or Tk::PlaceWindow?
+      my $x = int($w->screenwidth / 2 - $w->reqwidth / 2 - $w->parent->vrootx);
+      my $y = int($w->screenheight / 2 - $w->reqheight / 2 - $w->parent->vrooty);
+      $w->geometry("+$x+$y");
+      $w->deiconify;
+     }
+    else
+     {
+      $w->Popup;
+     }
 
-    {
-	my $title = $w->cget(-title);
-	if (!defined $title) {
-	    $title = ($w->cget(-type) eq 'open' ? 'Open' : 'Save As');
-	}
-	$w->title($title);
-    }
-
-    $w->deiconify;
     # Set a grab and claim the focus too.
+#XXX use Tk::setFocusGrab when it's available
     my $oldFocus = $w->focusCurrent;
     my $oldGrab = $w->grabCurrent;
     my $grabStatus = $oldGrab->grabStatus if ($oldGrab);
@@ -295,17 +335,18 @@ sub Show {
     my $ent = $w->{'ent'};
     $ent->focus;
     $ent->delete(0, 'end');
-    $ent->insert(0, $w->{'selectFile'});
-    $ent->selectionFrom(0);
-    $ent->selectionTo('end');
-    $ent->icursor('end');
+    if (defined $w->{'selectFile'} && $w->{'selectFile'} ne '') {
+	$ent->insert(0, $w->{'selectFile'});
+	$ent->selectionRange(0,'end');
+	$ent->icursor('end');
+    }
 
     # 8. Wait for the user to respond, then restore the focus and
     # return the index of the selected button.  Restore the focus
     # before deleting the window, since otherwise the window manager
     # may take the focus away so we can't redirect it.  Finally,
     # restore any grab that was in effect.
-    $w->waitVariable(\$selectFilePath);
+    $w->waitVariable(\$w->{'selectFilePath'});
     eval {
 	$oldFocus->focus if $oldFocus;
     };
@@ -320,15 +361,15 @@ sub Show {
 	    $oldGrab->grab;
 	}
     }
-    return $selectFilePath;
+    return $w->{'selectFilePath'};
 }
 
 # tkFDialog_UpdateWhenIdle --
 #
-#	Creates an idle event handler which updates the dialog in idle
-#	time. This is important because loading the directory may take a long
-#	time and we don't want to load the same directory for multiple times
-#	due to multiple concurrent events.
+#       Creates an idle event handler which updates the dialog in idle
+#       time. This is important because loading the directory may take a long
+#       time and we don't want to load the same directory for multiple times
+#       due to multiple concurrent events.
 #
 sub UpdateWhenIdle {
     my $w = shift;
@@ -339,38 +380,11 @@ sub UpdateWhenIdle {
     }
 }
 
-# This proc gets called whenever data(selectPath) is set
-#
-sub SetPath
-{
- my $w = shift;
- if (@_)
-  {
-   my $newdir = shift;
-   # Update the Directory: option menu
-   my @list;
-   my $dir = '';
-   foreach my $subdir (TclFileSplit($newdir))
-    {
-     $dir = TclFileJoin($dir, $subdir);
-     push @list, $dir;
-    }
-    my $dirMenu = $w->{'dirMenu'};
-    $dirMenu->configure(-options => \@list);
-    $w->{'selectPath'} = $newdir;
-    # my $vvar = $dirMenu->cget('-variable');
-    # my $tvar = $dirMenu->cget('-textvariable');
-    # my $mvar = \$w->{'selectPath'};
-    # print "now m=$mvar ($$mvar) t=$tvar ($$tvar) v=$vvar ($$vvar)\n";
-  }
- $w->UpdateWhenIdle;
-}
-
 # tkFDialog_Update --
 #
-#	Loads the files and directories into the IconList widget. Also
-#	sets up the directory option menu for quick access to parent
-#	directories.
+#       Loads the files and directories into the IconList widget. Also
+#       sets up the directory option menu for quick access to parent
+#       directories.
 #
 sub Update {
     my $w = shift;
@@ -378,7 +392,7 @@ sub Update {
 
     # This proc may be called within an idle handler. Make sure that the
     # window has not been destroyed before this proc is called
-    if (!Tk::Exists($w) || !$w->isa('Tk::FBox')) {
+    if (!Tk::Exists($w) || $w->class ne 'FBox') {
 	return;
     } else {
 	delete $w->{'updateId'};
@@ -391,14 +405,14 @@ sub Update {
     my $folder = $folderImage->{$w->MainWindow};
     my $file   = $fileImage->{$w->MainWindow};
     my $appPWD = _cwd();
-    if (!ext_chdir($w->{'selectPath'})) {
+    if (!ext_chdir($w->_get_select_path)) {
 	# We cannot change directory to $data(selectPath). $data(selectPath)
 	# should have been checked before tkFDialog_Update is called, so
 	# we normally won't come to here. Anyways, give an error and abort
 	# action.
 	$w->messageBox(-type => 'OK',
 		       -message => 'Cannot change to the directory "' .
-		       $w->{'selectPath'} . "\".\nPermission denied.",
+		       $w->_get_select_path . "\".\nPermission denied.",
 		       -icon => 'warning',
 		      );
 	ext_chdir($appPWD);
@@ -407,31 +421,59 @@ sub Update {
 
     # Turn on the busy cursor. BUG?? We haven't disabled X events, though,
     # so the user may still click and cause havoc ...
-    $w->Busy(-cursor => 'watch', -recurse => 1);
+    my $ent = $w->{'ent'};
+    my $entCursor = $ent->cget(-cursor);
+    my $dlgCursor = $w->cget(-cursor);
+    $ent->configure(-cursor => 'watch');
+    $w->configure(-cursor => 'watch');
     $w->idletasks;
     my $icons = $w->{'icons'};
     $icons->DeleteAll;
 
     # Make the dir & file list
-    my $flt = join('|', split(' ', $w->cget(-filter)) );
-    $flt =~ s!([\.\+])!\\$1!g;
-    $flt =~ s!\*!.*!g;
+    my $cwd = _cwd();
     local *FDIR;
-    if( opendir( FDIR,  _cwd() )) {
-        my @files;
-#	my $sortcmd = $w->cget(-sortcmd);
+    if (opendir(FDIR, $cwd)) {
+	my @files;
+#       my $sortcmd = $w->cget(-sortcmd);
 	my $sortcmd = sub { $w->cget(-sortcmd)->($a,$b) };
-        foreach my $f (sort $sortcmd readdir(FDIR)) {
-            next if $f eq '.' or $f eq '..';
-            if (-d $f) { $icons->Add($folder, $f); }
-            elsif( $f =~ m!$flt$! ) { push( @files, $f ); }
+	my $flt = $w->cget(-filter);
+	my $fltcb;
+	if (ref $flt eq 'CODE') {
+	    $fltcb = $flt;
+	} else {
+	    $flt = _rx_to_glob($flt);
 	}
-      closedir( FDIR );
-      foreach my $f ( @files ) { $icons->Add($file, $f); }
+	my $type_dir = $w->cget(-type) eq 'dir';
+	foreach my $f (sort $sortcmd readdir(FDIR)) {
+	    next if $f eq '.' or $f eq '..';
+	    next if $type_dir && ! -d "$cwd/$f"; # XXX use File::Spec?
+	    if ($fltcb) {
+		next if !$fltcb->($w, $f, $cwd);
+	    } else {
+		next if !-d $f && $f !~ m!$flt!;
+	    }
+	    if (-d $f) {
+		$icons->Add($folder, $f);
+	    } else {
+		push @files, $f;
+	    }
+	}
+	closedir(FDIR);
+	$icons->Add($file, @files);
     }
 
     $icons->Arrange;
 
+    # Update the Directory: option menu
+    my @list;
+    my $dir = '';
+    foreach my $subdir (TclFileSplit($w->_get_select_path)) {
+	$dir = TclFileJoin($dir, $subdir);
+	push @list, $dir;
+    }
+    my $dirMenu = $w->{'dirMenu'};
+    $dirMenu->configure(-options => \@list);
 
     # Restore the PWD to the application's PWD
     ext_chdir($appPWD);
@@ -442,12 +484,13 @@ sub Update {
     }
 
     # turn off the busy cursor.
-    $w->Unbusy;
+    $ent->configure(-cursor => $entCursor);
+    $w->configure(-cursor =>  $dlgCursor);
 }
 
 # tkFDialog_SetPathSilently --
 #
-# 	Sets data(selectPath) without invoking the trace procedure
+#       Sets data(selectPath) without invoking the trace procedure
 #
 sub SetPathSilently {
     my($w, $path) = @_;
@@ -455,8 +498,17 @@ sub SetPathSilently {
     $w->{'selectPath'} = $path;
 }
 
+# This proc gets called whenever data(selectPath) is set
+#
+sub SetPath {
+    my $w = shift;
+    $w->{'selectPath'} = $_[0] if @_;
+    $w->UpdateWhenIdle;
+}
+
 # This proc gets called whenever data(filter) is set
 #
+#XXX here's much more code in the tcl version ... check it out
 sub SetFilter {
     my($w, $title, $filter) = @_;
     $w->configure(-filter => $filter);
@@ -468,41 +520,46 @@ sub SetFilter {
 
 # tkFDialogResolveFile --
 #
-#	Interpret the user's text input in a file selection dialog.
-#	Performs:
+#       Interpret the user's text input in a file selection dialog.
+#       Performs:
 #
-#	(1) ~ substitution
-#	(2) resolve all instances of . and ..
-#	(3) check for non-existent files/directories
-#	(4) check for chdir permissions
+#       (1) ~ substitution
+#       (2) resolve all instances of . and ..
+#       (3) check for non-existent files/directories
+#       (4) check for chdir permissions
 #
 # Arguments:
-#	context:  the current directory you are in
-#	text:	  the text entered by the user
-#	defaultext: the default extension to add to files with no extension
+#       context:  the current directory you are in
+#       text:     the text entered by the user
+#       defaultext: the default extension to add to files with no extension
 #
 # Return value:
-#	[list $flag $directory $file]
+#       [list $flag $directory $file]
 #
-#	 flag = OK	: valid input
-#	      = PATTERN	: valid directory/pattern
-#	      = PATH	: the directory does not exist
-#	      = FILE	: the directory exists but the file doesn't
-#			  exist
-#	      = CHDIR	: Cannot change to the directory
-#	      = ERROR	: Invalid entry
+#        flag = OK      : valid input
+#             = PATTERN : valid directory/pattern
+#             = PATH    : the directory does not exist
+#             = FILE    : the directory exists but the file doesn't
+#                         exist
+#             = CHDIR   : Cannot change to the directory
+#             = ERROR   : Invalid entry
 #
-#	 directory      : valid only if flag = OK or PATTERN or FILE
-#	 file           : valid only if flag = OK or PATTERN
+#        directory      : valid only if flag = OK or PATTERN or FILE
+#        file           : valid only if flag = OK or PATTERN
 #
-#	directory may not be the same as context, because text may contain
-#	a subdirectory name
+#       directory may not be the same as context, because text may contain
+#       a subdirectory name
 #
 sub ResolveFile {
     my($context, $text, $defaultext) = @_;
     my $appPWD = _cwd();
     my $path = JoinFile($context, $text);
-    $path = "$path$defaultext" if ($path !~ /\..+$/) and defined $defaultext;
+    # If the file has no extension, append the default.  Be careful not
+    # to do this for directories, otherwise typing a dirname in the box
+    # will give back "dirname.extension" instead of trying to change dir.
+    if (!-d $path && $path !~ /\..+$/ && defined $defaultext) {
+	$path = "$path$defaultext";
+    }
     # Cannot just test for existance here as non-existing files are
     # not an error for getSaveFile type dialogs.
     # return ('ERROR', $path, "") if (!-e $path);
@@ -557,15 +614,14 @@ sub EntFocusIn {
     my $w = shift;
     my $ent = $w->{'ent'};
     if ($ent->get ne '') {
-	$ent->selectionFrom(0);
-	$ent->selectionTo('end');
+	$ent->selectionRange(0, 'end');
 	$ent->icursor('end');
     } else {
 	$ent->selectionClear;
     }
-    $w->{'icons'}->Unselect;
+#XXX is this missing in the tcl version, too???    $w->{'icons'}->Selection('clear');
     my $okBtn = $w->{'okBtn'};
-    if ($w->cget(-type) eq 'open') {
+    if ($w->cget(-type) ne 'save') {
 	$okBtn->configure(-text => 'Open');
     } else {
 	$okBtn->configure(-text => 'Save');
@@ -581,12 +637,32 @@ sub EntFocusOut {
 #
 sub ActivateEnt {
     my $w = shift;
-    my $ent = $w->{'ent'};
-    my $text = $ent->get;
-    $text =~ s/^\s+//;
-    $text =~ s/\s+$//;
-    my($flag, $path, $file) = ResolveFile($w->{'selectPath'}, $text,
+    if ($w->cget(-multiple)) {
+	# For the multiple case we have to be careful to get the file
+	# names as a true list, watching out for a single file with a
+	# space in the name.  Thus we query the IconList directly.
+
+	$w->{'selectFile'} = [];
+	for my $item ($w->{'icons'}->Curselection) {
+	    $w->VerifyFileName($w->_get_from_icons($item));
+	}
+    } else {
+	my $ent = $w->{'ent'};
+	my $text = $w->_encode_filename($ent->get);
+	$w->VerifyFileName($text);
+    }
+}
+
+# Verification procedure
+#
+sub VerifyFileName {
+    my($w, $text) = @_;
+#XXX leave this here?
+#    $text =~ s/^\s+//;
+#    $text =~ s/\s+$//;
+    my($flag, $path, $file) = ResolveFile($w->_get_select_path, $text,
 					  $w->cget(-defaultextension));
+    my $ent = $w->{'ent'};
     if ($flag eq 'OK') {
 	if ($file eq '') {
 	    # user has entered an existing (sub)directory
@@ -594,7 +670,11 @@ sub ActivateEnt {
 	    $ent->delete(0, 'end');
 	} else {
 	    $w->SetPathSilently($path);
-	    $w->{'selectFile'} = $file;
+	    if ($w->cget(-multiple)) {
+		push @{ $w->{'selectFile'} }, $file;
+	    } else {
+		$w->{'selectFile'} = $file;
+	    }
 	    $w->Done;
 	}
     } elsif ($flag eq 'PATTERN') {
@@ -606,34 +686,34 @@ sub ActivateEnt {
 			   -type => 'OK',
 			   -message => 'File "' . TclFileJoin($path, $file)
 			   . '" does not exist.');
-	    $ent->selection('from', 0);
-	    $ent->selection('to', 'end');
+	    $ent->selectionRange(0, 'end');
 	    $ent->icursor('end');
-	} else {
+	} elsif ($w->cget(-type) eq 'save') {
 	    $w->SetPathSilently($path);
-	    $w->{'selectFile'} = $file;
+	    if ($w->cget(-multiple)) {
+		push @{ $w->{'selectFile'} }, $file;
+	    } else {
+		$w->{'selectFile'} = $file;
+	    }
 	    $w->Done;
 	}
     } elsif ($flag eq 'PATH') {
 	$w->messageBox(-icon => 'warning',
 		       -type => 'OK',
 		       -message => "Directory \'$path\' does not exist.");
-	$ent->selection('from', 0);
-	$ent->selection('to', 'end');
+	$ent->selectionRange(0, 'end');
 	$ent->icursor('end');
     } elsif ($flag eq 'CHDIR') {
 	$w->messageBox(-type => 'OK',
 		       -message => "Cannot change to the directory \"$path\".\nPermission denied.",
 		       -icon => 'warning');
-	$ent->selection('from', 0);
-	$ent->selection('to', 'end');
+	$ent->selectionRange(0, 'end');
 	$ent->icursor('end');
     } elsif ($flag eq 'ERROR') {
 	$w->messageBox(-type => 'OK',
 		       -message => "Invalid file name \"$path\".",
 		       -icon => 'warning');
-	$ent->selection('from', 0);
-	$ent->selection('to', 'end');
+	$ent->selectionRange(0, 'end');
 	$ent->icursor('end');
     }
 }
@@ -650,8 +730,8 @@ sub InvokeBtn {
 #
 sub UpDirCmd {
     my $w = shift;
-    $w->SetPath(File::Basename::dirname($w->{'selectPath'}))
-      unless ($w->{'selectPath'} eq '/');
+    $w->SetPath(File::Basename::dirname($w->_get_select_path))
+      unless ($w->_get_select_path eq '/');
 }
 
 # Join a file name to a path name. The "file join" command will break
@@ -711,36 +791,70 @@ sub TclFileSplit {
 #
 sub OkCmd {
     my $w = shift;
-    my $text = $w->{'icons'}->Get;
-    if (defined $text and $text ne '') {
-	my $file = JoinFile($w->{'selectPath'}, $text);
+    my $from = shift || "button";
+
+    my $filenames = [];
+    for my $item ($w->{'icons'}->Curselection) {
+	push @$filenames, $w->_get_from_icons($item);
+    }
+
+    my $filename = $filenames->[0];
+    $filename = "" if !defined $filename;
+    if ($w->cget('-type') eq 'dir' && $from ne "iconlist") {
+	my $file = $filename eq '' ? $w->_get_select_path : JoinFile($w->_get_select_path, $filename);
+	$w->Done($file);
+    } elsif ((@$filenames && !$w->cget('-multiple')) ||
+	($w->cget('-multiple') && @$filenames == 1)) {
+	my $file = JoinFile($w->_get_select_path, $filename);
 	if (-d $file) {
-	    $w->ListInvoke($text);
+	    $w->ListInvoke($filename);
 	    return;
 	}
     }
+
     $w->ActivateEnt;
 }
 
 # Gets called when user presses the "Cancel" button
 #
 sub CancelCmd {
-    undef $selectFilePath;
+    my $w = shift;
+    undef $w->{'selectFilePath'};
 }
 
 # Gets called when user browses the IconList widget (dragging mouse, arrow
 # keys, etc)
 #
 sub ListBrowse {
-    my($w, $text) = @_;
-    return if ($text eq '');
-    my $file = JoinFile($w->{'selectPath'}, $text);
+    my($w) = @_;
+
+    my $text = [];
+    for my $item ($w->{'icons'}->Curselection) {
+	push @$text, $w->_get_from_icons($item);
+    }
+    return if @$text == 0;
+    my $isDir;
+    if (@$text > 1) {
+	my $newtext = [];
+	for my $file (@$text) {
+	    my $fullfile = JoinFile($w->_get_select_path, $file);
+	    if (!-d $fullfile) {
+		push @$newtext, $file;
+	    }
+	}
+	$text = $newtext;
+	$isDir = 0;
+    } else {
+	my $file = JoinFile($w->_get_select_path, $text->[0]);
+	$isDir = -d $file;
+    }
     my $ent = $w->{'ent'};
     my $okBtn = $w->{'okBtn'};
-    unless (-d $file) {
-	$ent->delete(0, 'end');
-	$ent->insert(0, $text);
-	if ($w->cget(-type) eq 'open') {
+    if (!$isDir) {
+	$ent->delete(qw(0 end));
+	$ent->insert(0, "@$text"); # XXX quote!
+
+	if ($w->cget('-type') ne 'save') {
 	    $okBtn->configure(-text => 'Open');
 	} else {
 	    $okBtn->configure(-text => 'Save');
@@ -754,9 +868,9 @@ sub ListBrowse {
 # Return key, etc)
 #
 sub ListInvoke {
-    my($w, $text) = @_;
-    return if ($text eq '');
-    my $file = JoinFile($w->{'selectPath'}, $text);
+    my($w, @filenames) = @_;
+    return if !@filenames;
+    my $file = JoinFile($w->_get_select_path, $filenames[0]);
     if (-d $file) {
 	my $appPWD = _cwd();
 	if (!ext_chdir($file)) {
@@ -768,48 +882,55 @@ sub ListInvoke {
 	    $w->SetPath($file);
 	}
     } else {
-        my($flag, $path, $file) = ResolveFile($w->{'selectPath'}, $text);
-        if ($flag ne 'OK') {
-            $w->messageBox(-type => 'OK',
-                           -message => "Cannot resolve $w->{'selectPath'}/$text.",
-                           -icon => 'error');
-        } else {
-            $path = JoinFile($path, $file);
-            $w->Done($path);
-        }
+	if ($w->cget('-multiple')) {
+	    $w->{'selectFile'} = [@filenames];
+	} else {
+	    $w->{'selectFile'} = $file;
+	}
+	$w->Done;
     }
 }
 
 # tkFDialog_Done --
 #
-#	Gets called when user has input a valid filename.  Pops up a
-#	dialog box to confirm selection when necessary. Sets the
-#	tkPriv(selectFilePath) variable, which will break the "tkwait"
-#	loop in tkFDialog and return the selected filename to the
-#	script that calls tk_getOpenFile or tk_getSaveFile
+#       Gets called when user has input a valid filename.  Pops up a
+#       dialog box to confirm selection when necessary. Sets the
+#       tkPriv(selectFilePath) variable, which will break the "tkwait"
+#       loop in tkFDialog and return the selected filename to the
+#       script that calls tk_getOpenFile or tk_getSaveFile
 #
 sub Done {
     my $w = shift;
-    my $_selectFilePath = (@_) ? shift : '';
-    if ($_selectFilePath eq '') {
-	$_selectFilePath = JoinFile($w->{'selectPath'}, $w->{'selectFile'});
-	if (-e $_selectFilePath and
-	    $w->cget(-type) eq 'save' and
+    my $selectFilePath = (@_) ? shift : '';
+    if ($selectFilePath eq '') {
+	if ($w->cget('-multiple')) {
+	    $selectFilePath = [];
+	    for my $f (@{ $w->{'selectFile'} }) {
+		push @$selectFilePath, JoinFile($w->_get_select_path, $f);
+	    }
+	} else {
+	    $selectFilePath = JoinFile($w->_get_select_path,
+				       $w->{'selectFile'});
+	}
+	if ($w->cget(-type) eq 'save' and
+	    -e $selectFilePath and
 	    !$w->cget(-force)) {
 	    my $reply = $w->messageBox
 	      (-icon => 'warning',
 	       -type => 'YesNo',
-	       -message => "File \"$_selectFilePath\" already exists.\nDo you want to overwrite it?");
+	       -message => "File \"$selectFilePath\" already exists.\nDo you want to overwrite it?");
 	    return unless (lc($reply) eq 'yes');
 	}
     }
-    $selectFilePath = ($_selectFilePath ne '' ? $_selectFilePath : undef);
+    $w->{'selectFilePath'} = ($selectFilePath ne '' ? $selectFilePath : undef);
 }
 
 sub FDialog {
     my $cmd = shift;
     if ($cmd =~ /Save/) {
 	push @_, -type => 'save';
+    } elsif ($cmd =~ /Directory/) {
+        push @_, -type => 'dir';
     }
     Tk::DialogWrapper('FBox', $cmd, @_);
 }
@@ -824,10 +945,10 @@ sub GetFileTypes {
     my $in = shift;
     my %fileTypes;
     foreach my $t (@$in) {
-        if (@$t < 2  || @$t > 3) {
+	if (@$t < 2  || @$t > 3) {
 	    require Carp;
 	    Carp::croak("bad file type \"$t\", should be \"typeName [extension ?extensions ...?] ?[macType ?macTypes ...?]?\"");
-        }
+	}
 	push @{ $fileTypes{$t->[0]} }, (ref $t->[1] eq 'ARRAY'
 					? @{ $t->[1] }
 					: $t->[1]);
@@ -837,27 +958,27 @@ sub GetFileTypes {
     my %hasDoneType;
     my %hasGotExt;
     foreach my $t (@$in) {
-        my $label = $t->[0];
-        my @exts;
+	my $label = $t->[0];
+	my @exts;
 
-        next if (exists $hasDoneType{$label});
+	next if (exists $hasDoneType{$label});
 
-        my $name = "$label (";
+	my $name = "$label (";
 	my $sep = '';
-        foreach my $ext (@{ $fileTypes{$label} }) {
-            next if ($ext eq '');
-            $ext =~ s/^\./*./;
-            if (!exists $hasGotExt{$label}->{$ext}) {
-                $name .= "$sep$ext";
-                push @exts, $ext;
-                $hasGotExt{$label}->{$ext}++;
-            }
-            $sep = ',';
-        }
-        $name .= ')';
-        push @types, [$name, \@exts];
+	foreach my $ext (@{ $fileTypes{$label} }) {
+	    next if ($ext eq '');
+	    $ext =~ s/^\./*./;
+	    if (!exists $hasGotExt{$label}->{$ext}) {
+		$name .= "$sep$ext";
+		push @exts, $ext;
+		$hasGotExt{$label}->{$ext}++;
+	    }
+	    $sep = ',';
+	}
+	$name .= ')';
+	push @types, [$name, \@exts];
 
-        $hasDoneType{$label}++;
+	$hasDoneType{$label}++;
     }
 
     return @types;
@@ -910,6 +1031,33 @@ sub _untaint {
     $1;
 }
 
+sub _rx_to_glob {
+    my $arg = shift;
+    $arg = join('|', split(' ', $arg));
+    $arg =~ s!([\.\+])!\\$1!g;
+    $arg =~ s!\*!.*!g;
+    $arg = "^" . $arg . "\$";
+    if ($] >= 5.005) {
+	$arg = qr/$arg/;
+    }
+    $arg;
+}
+
+sub _get_from_icons {
+    my($w, $item) = @_;
+    $w->_encode_filename($w->{'icons'}->Get($item));
+}
+
+sub _get_select_path {
+    my($w) = @_;
+    $w->_encode_filename($w->{'selectPath'});
+}
+
+sub _encode_filename {
+    my($w, $filename) = @_;
+    $filename = $w->{encoding}->encode($filename);
+    $filename;
+}
+
 1;
 
-
